# -*- python -*-
# $RCSfile: core.py,v $
# $Revision: 1.8 $
# $Author: langer $
# $Date: 2007-08-10 21:24:21 $

## This file contains the API for gtk_logger.  Nothing outside of this
## file should have to be explicitly called by users.

import gtk
import string
import types
import weakref

import loggers
import logutils

_allexceptions = [Exception]

def start(filename, debugLevel=2, suppress_motion_events=True):
    global _suppress_motion_events
    logutils.set_debugLevel(debugLevel)
    _suppress_motion_events = suppress_motion_events
    if logutils.recording():
        logutils.logfile().close()
    try:
        logutils.set_logfile(open(filename, "w"))
    except:
        logutils.set_logfile(None)
        raise

def stop():
    if logutils.recording():
        try:
            logutils.logfile().close()
        finally:
            logutils.set_logfile(None)


# Utility functions to temporarily turn on motion-event recording.

def log_motion_events():
    logutils.log_motion_events()

def dont_log_motion_events():
    logutils.dont_log_motion_events()

def suppress_motion_events():
    return logutils.suppress_motion_events()

def add_exception(excclass):
    logutils.add_exception(excclass)

##################################

# Any gtk.Widget that needs to have its signals logged and replayed
# must have a name, and enough of its widget ancestors must have names
# that the sequence of names uniquely identifies the widget.
# setWidgetName is used to assign the names.

def setWidgetName(widget, name):
    return logutils.setWidgetName(widget, name)

# Top-level widgets (mostly gtk.Windows) must have their names
# assigned by newTopLevelWidget() instead of setWidgetName().

def newTopLevelWidget(widget, name):
    return logutils.newTopLevelWidget(widget, name)

# gtk.GObjects which aren't gtk.Widgets but still need to have their
# signals logged should be registered with adoptGObject(). 'object' is
# the object being registered.  'parent' is an actual gtk.Widget, and
# 'access_method' is a method of the parent class that retrieves the
# adopted GObject.  For example, to make the gtk.Adjustment that's
# part of a gtk.HRange loggable, call
#    adoptGObject(range.get_adjustment(), range, range.get_adjustment)
# If there's no such parent class method, then use access_function
# instead of access_method.  The function will be called with the
# parent widget as its first argument.  In either case, additional
# arguments to the function can be passed in as access_args or
# access_kwargs.

def adoptGObject(object, parent, access_method=None, access_function=None,
                 access_args=(), access_kwargs={}):
    assert not hasattr(object, 'oofparent')
    assert parent is not None
    assert access_method is not None or access_function is not None
    object.oofparent = parent
    if access_method is not None:
        object.oofparent_access_method = access_method.__name__
    else:
        object.oofparent_access_function = access_function.__name__
    object.oofparent_access_args = access_args
    object.oofparent_access_kwargs = access_kwargs

# set_submenu should be used instead of gtk.MenuItem.set_submenu.

def set_submenu(menuitem, submenu):
    menuitem.set_submenu(submenu)
    submenu.oofparent = menuitem

# To have a gtk.GObject's actions logged, use gtklogger.connect or
# gtklogger.connect_after instead of gtk.GObject.connect.  If a widget
# that otherwise would have no callback needs to be logged, use
# gtklogger.connect_passive.  All of these functions install
# signalLogger as a callback.

def connect(object, signal, callback, *args):
    return GUISignals(
        object,
        # Because the callbacks for different signals have different
        # numbers of arguments, and because keyword arguments don't
        # work for gtk signal handlers, but signalLogger *does* need
        # to know what signal it's handling, we *append* the signal to
        # the arguments.  That way, we can always extract it inside
        # signalLogger.
        object.connect(signal, loggers.signalLogger, *(args + (signal,))),
        object.connect(signal, callback, *args)
        )

def connect_after(object, signal, callback, *args):
    return GUISignals(
        object,
        object.connect_after(signal, loggers.signalLogger,
                             *(args + (signal,))),
        object.connect_after(signal, callback, *args)
        )

# connect_passive should be used when an action needs to be logged and
# replayed, but wouldn't otherwise have a callback.  For example,
# keypresses in a gtk.Entry often aren't connected, but still need to
# be logged.

def connect_passive(object, signal, *args):
    return GUISignals(
        object,
        object.connect(signal, loggers.signalLogger, *(args + (signal,)))
        )

def connect_passive_after(object, signal, *args):
    return GUISignals(
        object,
        object.connect_after(signal, loggers.signalLogger, *(args + (signal,)))
        )

# Because connect() connects *two* signals instead of one, it would be
# a pain to keep track of them both for blocking and unblocking.  So
# instead connect() returns a GUISignals object, which holds a bunch
# of signals and can block and unblock them simultaneously.

class GUISignals:
    def __init__(self, widget, *signals):
        self.signals = signals
        self.widget = weakref.ref(widget)
    def block(self):
        map(self.widget().handler_block, self.signals)
    def unblock(self):
        map(self.widget().handler_unblock, self.signals)
    def block_log(self):
        self.widget().handler_block(self.signals[0])
    def unblock_log(self):
        self.widget().handler_unblock(self.signals[0])
    def disconnect(self):
        map(self.widget().disconnect, self.signals)
        self.signals = ()

# We need to be able to tell if a dialog is running (see
# GUILogLineRunner in replay.py).  We do that by using a modified
# Dialog class that keeps track of how many dialogs are open.
        
class Dialog(gtk.Dialog):
    def __init__(self, *args, **kwargs):
        super(Dialog, self).__init__(*args, **kwargs)
        connect_passive(self, 'delete-event')
        connect_passive(self, 'configure-event')
    def run(self):
        logutils.increment_dialog_level()
        try:
            result = gtk.Dialog.run(self)
        finally:
            logutils.decrement_dialog_level()
        return result
    def add_button(self, name, id):
        # Just to be nice to the programmer, turn on logging for the
        # buttons automatically.  Perhaps this should be optional.
        button = super(Dialog, self).add_button(name, id)
        assert type(name) is types.StringType
        setWidgetName(button, name)
        connect_passive(button, 'clicked')
        return button

# logScrollBars should be called on any ScrolledWindow whose scroll
# bars need to be logged.  It just encapsulates the adoptGObject calls
# that would otherwise be necessary.  It returns a pair of GUISignals
# objects, one for each scroll bar.  The return value can probably be
# ignored in most cases.

def logScrollBars(window, name=None):
    assert isinstance(window, gtk.ScrolledWindow)
    if name is not None:
        setWidgetName(window, name)
    adoptGObject(window.get_hadjustment(), window,
                 access_method=window.get_hadjustment)
    adoptGObject(window.get_vadjustment(), window,
                 access_method=window.get_vadjustment)
    return (
        connect_passive(window.get_hadjustment(), 'value-changed'),
        connect_passive(window.get_vadjustment(), 'value-changed'))



## Test for uniqueness by finding all named widgets and checking that
## findWidget(gtkPath(widget))==widget.

def sanity_check(widget, path=[]):
    widgetname = logutils.getWidgetName(widget)
    if widgetname:
        path = path + [widgetname]
        if findWidget(string.join(path, ':')) is not widget:
            raise ValueError("%s is not a unique widget name" % path)
    if isinstance(widget, gtk.Container):
        for child in widget.get_children():
            sanity_check(child, path)

def comprehensive_sanity_check():
    for widget in logutils.topwidgets.values():
        sanity_check(widget)
    
