# -*- python -*-
# $RCSfile: logutils.py,v $
# $Revision: 1.4 $
# $Author: langer $
# $Date: 2007-07-12 19:25:57 $

# Some of this stuff might want to live elsewhere, eventually... It's
# stuff that has been moved out of gtklogger.py to keep the public API
# file clean.

# This file also contains some global variables and access functions
# for them (so they're not really global variables...).  This helps to
# avoid import loops, because this files doesn't import anything else.

import gtk
import string
import weakref

import checkpoint


class GtkLoggerException(Exception):
    def __init__(self, *args):
        self.msg = string.join(args, ' ')
    def __str__(self):
        return self.msg

#############################
    
# Any gtk.Widget that needs to have its signals logged and replayed
# must have a name, and enough of its widget ancestors must have names
# that the sequence of names uniquely identifies the widget.
# setWidgetName is used to assign the names.

def setWidgetName(widget, name):
    assert isinstance(widget, gtk.Widget)
    assert name is not None
    # When saved in the log files, widget names are surrounded by
    # single quotation marks, so we have to make sure to use only
    # double quotes in widget names.  Since colons are used as
    # delimiters in widget paths, they must be replaced too, when they
    # occur in widget names.
    safename = name.replace("'", '"').replace(":", ";")
    
    widget.oofname = safename
    return safename

def getWidgetName(widget):
    return getattr(widget, 'oofname', None)

##############################

# Dictionary of top level widgets keyed by name
topwidgets = weakref.WeakValueDictionary()

class GtkLoggerTopFailure(GtkLoggerException):
    pass

def getTopLevelWidget(name):
    try:
        return topwidgets[name]
    except KeyError:
        raise GtkLoggerTopFailure(name)

def isTopLevelWidget(object):
    return object in topwidgets.values()

def getTopWidgetNames():
    return topwidgets.keys()[:]

# Top-level widgets (mostly gtk.Windows) must have their names
# assigned by newTopLevelWidget() instead of setWidgetName().

def newTopLevelWidget(widget, name):
    safename = setWidgetName(widget, name)
    widget.connect('map', _topmapCallback, name)
    topwidgets[safename] = widget

def _topmapCallback(object, name):
    checkpoint.checkpoint("toplevel widget mapped %s" % name)

##############################

## These findXXXX functions are mostly used within log files, but they
## can be useful within test functions as well. They're defined here
## so that test modules don't have to import replay.py.  replay.py
## imports the functions from here.

def findWidget(path):
    splitpath = path.split(':')
    top = getTopLevelWidget(splitpath[0])
    if len(splitpath) == 1:
        return top
    return _recursiveFindWidget(splitpath[1:], top)

# Given a path (list of widget names) and a base (starting widget),
# _recursiveFindGObject looks for and returns the Widget (or GObject)
# named by the path.  The search is a depth first search of the widget
# tree.
def _recursiveFindWidget(path, base):
    if not path:
        return base
    if isinstance(base, gtk.Container):
        for child in base.get_children():
            childname = getWidgetName(child)
            if childname == path[0]:
                return _recursiveFindWidget(path[1:], child)
            if childname is None:
                result = _recursiveFindWidget(path, child)
                if result is not None:
                    return result
                # Didn't find anything using unnamed widgets.  Keep
                # looking at other children.

def findMenu(widget, path):
    # widget is a gtk.MenuShell (container for MenuItems)
    splitpath = path.split(':', 1)
    for child in widget.get_children(): # child is a gtk.MenuItem
        if getWidgetName(child) == splitpath[0]:
            if len(splitpath) == 1:
                return child
            return findMenu(child.get_submenu(), splitpath[1])

def findCellRenderer(treeview, col, rend):
    return treeview.get_column(col).get_cell_renderers()[rend]

# findAllWidgets returns the paths to all named widgets under (and
# including) the given widget.  It *won't* find adopted non-widget
# objects.  It's mostly for debugging.

def findAllWidgets(topname):
    return _findAllWidgets(findWidget(topname))
def _findAllWidgets(top):
    topname = getWidgetName(top)
    childpaths = []
    if isinstance(top, gtk.Container):
        for child in top.get_children():
            childpaths.extend(_findAllWidgets(child))
    if topname:
        childpaths = [topname] + \
                     [string.join([topname, path], ':') for path in childpaths]
    return childpaths

##############################

# Keep track of the run level, which is the gtk run level plus the
# number of open modal dialogs.  See the comment in
# GUILogLineRunner.__call__.

dialog_level = 0
def run_level():
    return gtk.main_level() + dialog_level
def increment_dialog_level():
    global dialog_level
    dialog_level += 1

def decrement_dialog_level():
    global dialog_level
    dialog_level -= 1

###

_replaying = False
def replaying():
    return _replaying

def set_replaying(val):
    global _replaying
    _replaying = val

###

_logfile = None

def recording():
    return _logfile is not None

def logfile():
    return _logfile

def set_logfile(fileobj):
    global _logfile
    _logfile = fileobj


###

_debugLevel = 0

def debugLevel():
    return _debugLevel

def set_debugLevel(n):
    global _debugLevel
    _debugLevel = n

         
###

_suppress_motion_events = True

def log_motion_events():
    global _suppress_motion_events
    _suppress_motion_events = False

def dont_log_motion_events():
    global _suppress_motion_events
    _suppress_motion_events = True

def suppress_motion_events():
    return _suppress_motion_events

###

_allexceptions = [Exception]

def add_exception(excclass):
    _allexceptions.append(excclass)

def exceptions():
    return tuple(_allexceptions)
