// -*- C++ -*-
// $RCSfile: ooferror.h,v $
// $Revision: 1.14 $
// $Author: reida $
// $Date: 2006-05-08 20:30:05 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#ifndef OOFERROR_H
#define OOFERROR_H

#include <string>
#include <Python.h>

// Error namespace causes too many problems for SWIG.  Instead,
// use a customizing prefix for error names.

// Base class so that all OOF errors can be caught by the one catch
// statement typemap in oof.swg.  Also holds the One True Pythonizer
// for error classes.

class ErrError {
public:
  static PyObject *pyconverter;
  virtual ~ErrError() {}
  virtual std::string pythonequiv() const = 0;
  virtual std::string message() const = 0;
  virtual void throw_self() const = 0;
};

// ErrErrorBase exists just so that a base class pointer can re-throw
// itself.  It's used in pythonErrorRelay, when throwing a C++
// exception that was originally raised in Python.  The template
// trickery is necessary because it's not possible to throw an object
// of an abstract class (such as ErrError).  The template parameter
// must be the derived class.

template <class E>
class ErrErrorBase : public ErrError {
public:
  virtual void throw_self() const {
    const E *self = dynamic_cast<const E*>(this);
    throw *self;
  }
};

void pyErrorInit(PyObject*);


// Programming errors are fatal (to the program...)

template <class E>
class ErrProgrammingErrorBase : public ErrErrorBase<E> {
protected:
  const std::string file;
  const int line;
  const std::string msg;
public:
  ErrProgrammingErrorBase(const std::string &f, int l)
    : file(f), line(l), msg("")
  {}
  ErrProgrammingErrorBase(const std::string &m, const std::string &f, int l)
    : file(f), line(l), msg(m)
  {}
  virtual ~ErrProgrammingErrorBase() {}
  virtual std::string message() const { return msg; }
  std::string filename() const { return file; } // For reading from Python.
  int lineno() const { return line; }           // ... this one too.
};

class ErrProgrammingError
  : public ErrProgrammingErrorBase<ErrProgrammingError>
{
public:
  ErrProgrammingError(const std::string &f, int l)
    : ErrProgrammingErrorBase<ErrProgrammingError>(f, l)
  {}
  ErrProgrammingError(const std::string &m, const std::string &f, int l)
    : ErrProgrammingErrorBase<ErrProgrammingError>(m, f, l)
  {}
  virtual std::string pythonequiv() const;
};


// Resource shortages are probably fatal.

class ErrResourceShortage : public ErrErrorBase<ErrResourceShortage> {
private:
  const std::string msg;
public:
  ErrResourceShortage(const std::string &m) 
    : msg(m) 
  {}
  virtual ~ErrResourceShortage() {}
  virtual std::string message() const { return msg; }
  virtual std::string pythonequiv() const;
};


class ErrBoundsError : public ErrErrorBase<ErrBoundsError> {
private:
  const std::string msg;
public:
  ErrBoundsError(const std::string &m) 
    : msg(m) 
  {}
  virtual ~ErrBoundsError() {}
  virtual std::string message() const { return msg; }
  virtual std::string pythonequiv() const;
};




class ErrBadIndex : public ErrProgrammingErrorBase<ErrBadIndex> {
private:
  int badindex;
public:
  ErrBadIndex(int i, const std::string &f, int l)
    : ErrProgrammingErrorBase<ErrBadIndex>(f, l),
      badindex(i)
  {}
  virtual std::string message() const;
  virtual std::string pythonequiv() const;
};


// User errors shouldn't be fatal.

template <class E>
class ErrUserErrorBase : public ErrErrorBase<E> {
public:
  const std::string msg;
  ErrUserErrorBase(const std::string &m) : msg(m) {}
  virtual ~ErrUserErrorBase() {}
  virtual std::string message() const { return msg; }
  // pythonequiv must be defined in derived classes.
};

// Generic user error
class ErrUserError : public ErrUserErrorBase<ErrUserError> {
public:
  ErrUserError(const std::string &m) : ErrUserErrorBase<ErrUserError>(m) {}
  virtual std::string pythonequiv() const;
};

class ErrSetupError : public ErrUserErrorBase<ErrSetupError> {
public:
  ErrSetupError(const std::string &m) : ErrUserErrorBase<ErrSetupError>(m) {}
  virtual std::string pythonequiv() const;
};


class ErrProcessAborted :public ErrErrorBase<ErrProcessAborted> {
private:
  const std::string pname;
public:
  ErrProcessAborted(const std::string &process) : pname(process) {}
  virtual ~ErrProcessAborted() {}
  virtual std::string pythonequiv() const;
  virtual std::string message() const { return pname; }
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Separate "stub" class for passing Python errors out through C++.
// This happens when a callback sets the Python error state, but 
// doesn't return to the Python environment.  The C++ caller should
// detect the error (by NULL return), and throw this exception,
// which will be caught in the exception typemap.

class PythonError {};

// pythonErrorRelay() should be called when C++ detects that a Python
// exception has been raised in a Python API call, if there's any
// chance that the exception might want to be handled in C++.  It
// converts the Python exception to a C++ exception (ie, an ErrError
// subclass) if possible.  Otherwise it just throws PythonError.

void pythonErrorRelay();

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void testerror();

#endif
