// -*- C++ -*-
// $RCSfile: field.C,v $
// $Revision: 1.41 $
// $Author: reida $
// $Date: 2007-04-20 19:14:26 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "common/removeitem.h"
#include "common/switchboard.h"
#include "common/tostring.h"
#include "common/trace.h"
#include "engine/csubproblem.h"
#include "engine/elementnodeiterator.h"
#include "engine/femesh.h"
#include "engine/field.h"
#include "engine/fieldeqnlist.h"
#include "engine/indextypes.h"
#include "engine/meshiterator.h"
#include "engine/node.h"
#include "engine/ooferror.h"
#include "engine/outputval.h"
#include "engine/property.h"
#include <iostream>
#include <vector>

std::vector<Field*> &Field::all() {
  static std::vector<Field*> all_fields;
  return all_fields;
}

std::vector<CompoundField*> &CompoundField::allcompoundfields() {
  static std::vector<CompoundField*> all_fields;
  return all_fields;
}

Field *getFieldByIndex(int index) {
  return Field::all()[index];
}

CompoundField *getCompoundFieldByIndex(int index) {
  return CompoundField::allcompoundfields()[index];

}

FEWrapper<Field>::AllWrappers
&Field::FindAllFieldWrappers::operator()()
{
  return mesh->fieldwrappers;
}

bool operator<(const Field::FieldData &a, const Field::FieldData &b) {
  if(a.active == b.active)
    return dynamic_cast<const FieldEqnData&>(a) < 
      dynamic_cast<const FieldEqnData&>(b);
  return a.active < b.active;
}

int countFields() {
  return Field::all().size(); 
}

int countCompoundFields() {
  return CompoundField::allcompoundfields().size();
}

Field *Field::getField(const std::string &nm) {
  const std::vector<Field*> &list = all();
  for(std::vector<Field*>::size_type i=0; i<list.size(); i++) {
    if(list[i]->name() == nm) {
      return list[i];
    }
  }
  throw ErrProgrammingError("Unknown Field \"" + nm + "\"",
			    __FILE__, __LINE__);
}

Field::Field(const std::string &nm, int dofs)
  : name_(nm),
    index_(all().size()),
    in_plane_(true),
    dim(dofs)
{
  all().push_back(this);
}


Field::~Field() {
  // this should be called only if the program is exiting... Fields
  // are global static objects and aren't deleted
  remove_item(all(), this);
}

std::ostream &operator<<(std::ostream &os, const Field &field) {
  os << field.name();
  return os;
}

const std::string &Field::name() const {
  return name_;
}

//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//

void Field::define(CSubProblem *subproblem) {
  subproblem->do_define_field(*this);
}

void Field::undefine(CSubProblem *subproblem) {
  subproblem->do_undefine_field(*this);
}

bool Field::is_defined(const CSubProblem *subproblem) const {
  return subproblem->is_defined_field(*this);
}

// initialization of a Field with a C++ function
void Field::initialize(CSubProblem *subproblem,
		       double (*f)(const Coord &position, int dof))
{
  subproblem->define_field(*this);
  for(FuncNodeIterator i=subproblem->funcnode_iterator(); !i.end(); ++i) {
    for(int d = 0; d<dim; ++d) {
      operator()(*i.node(), d)->value() = (*f)(i.node()->position(), d);
    }
  }
}

// initialization of a Field with a Python function
void Field::initialize(CSubProblem *subproblem,
		       double (*f)(const Coord &, int, void *),
		       void *clientdata)
{
  subproblem->define_field(*this);
  for(FuncNodeIterator i=subproblem->funcnode_iterator(); !i.end(); ++i) {
    for(int d = 0; d<dim; ++d) {
      operator()(*i.node(), d)->value() = (*f)(i.node()->position(), d,
					       clientdata);
    }
  }
}

//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//-\\-//

void Field::activate(CSubProblem *subproblem) {
  subproblem->do_activate_field(*this);
}

void Field::deactivate(CSubProblem *subproblem) {
  subproblem->do_deactivate_field(*this);
}

bool Field::is_active(const CSubProblem *subproblem) const {
  return subproblem->is_active_field(*this);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Field::registerProperty(Property *prop) const {
  prop->require_field(*this);
}

void CompoundField::registerProperty(Property *prop) const {
  prop->require_field(*this);
  prop->require_field(*zfield_);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// position of a given component in the dof lists in the nodes

int Field::localindex(const FuncNode *node, const FieldIndex &component) const {
  return node->fieldset.offset(this) + component.integer();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

double Field::value(const FuncNode *node, int component) const {
  return operator()(node, component)->value();
}

double Field::value(const ElementFuncNodeIterator &node, int component) const {
  return operator()(node, component)->value();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

CompoundField::CompoundField(const std::string &name, int dim, Field *follies)
  : Field(name, dim),
    zfield_(follies),
    cfield_indx(allcompoundfields().size())
{
  zfield_->set_oop(); // Set z component to be out-of-plane.
  allcompoundfields().push_back(this);
}

CompoundField::~CompoundField() {
  remove_item(allcompoundfields(), this);
}

// void CompoundField::set_in_plane(FEMesh *mesh, bool planarity) {
//   mesh->set_in_plane(*this, planarity);
//   // activate the out of plane field only if this field is active and
//   // not in plane.
//   if(in_plane(mesh))
//     subproblem->do_deactivate_field(*zfield_);
//   else {
//     if(subproblem->is_active_field(*this)) {
//       subproblem->do_activate_field(*zfield_);
//     }
//   }
// }

bool CompoundField::in_plane(const FEMesh *mesh) const {
  return mesh->in_plane(*this);
}

bool CompoundField::in_plane(const CSubProblem *subproblem) const {
  return subproblem->mesh->in_plane(*this);
}

void CompoundField::define(CSubProblem *subproblem) {
  subproblem->do_define_field(*this);
  subproblem->do_define_field(*zfield_);
}

void CompoundField::undefine(CSubProblem *subproblem) {
  subproblem->do_undefine_field(*this);
  subproblem->do_undefine_field(*zfield_);
}

void CompoundField::activate(CSubProblem *subproblem) {
  subproblem->do_activate_field(*this);
  if(!in_plane(subproblem->mesh))
    subproblem->do_activate_field(*zfield_);
}

void CompoundField::deactivate(CSubProblem *subproblem) {
  subproblem->do_deactivate_field(*this);
  subproblem->do_deactivate_field(*zfield_);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

ScalarField::ScalarField(const std::string &nm)
  : Field(nm, 1),
    ScalarFieldBase(nm),
    CompoundField(nm, 1, new ScalarFieldBase(nm+std::string("_z")))
{
}

const std::string ScalarField::classname_("ScalarField");

DegreeOfFreedom *ScalarFieldBase::operator()(const FuncNode *node) const {
  assert(node->hasField(*this));
  return node->doflist[node->fieldset.offset(this)];
}

DegreeOfFreedom *ScalarFieldBase::operator()(const ElementFuncNodeIterator &ei)
  const 
{
  return operator()(ei.funcnode());
}

// These two silly versions have to exist because these two functions
// have to be virtual in the Field base class. That's because
// CompoundField::out_of_plane returns a pointer to a Field base class
// object.
DegreeOfFreedom *ScalarFieldBase::operator()(const FuncNode *node,
					     int component) const
{
  assert(component == 0);
  assert(node->hasField(*this));
  return node->doflist[node->fieldset.offset(this)];
}

DegreeOfFreedom *ScalarFieldBase::operator()(const ElementFuncNodeIterator &ei,
					     int component) const 
{
  assert(component == 0);
  return operator()(ei.funcnode());
}

OutputValue ScalarFieldBase::newOutputValue() const {
  return OutputValue(new ScalarOutputVal(0.));
}

OutputValue ScalarFieldBase::output(const FuncNode &node) const {
  if(node.hasField(*this))
    return OutputValue(new ScalarOutputVal(operator()(node)->value()));
  else
    return OutputValue(new ScalarOutputVal(0.0));
}

OutputValue ScalarFieldBase::output(const ElementFuncNodeIterator &node) const {
  if(node.hasField(*this))
    return OutputValue(new ScalarOutputVal(operator()(node)->value()));
  else
    return OutputValue(new ScalarOutputVal(0.0));
}

void ScalarFieldBase::setValueFromOutputValue(const FuncNode &node,
					      const OutputValue *ov)
{
  assert(node.hasField(*this));
  const ScalarOutputVal &sov =
    dynamic_cast<const ScalarOutputVal&>(ov->valueRef());
  operator()(node)->value() = sov.value();
}


IteratorP ScalarFieldBase::iterator(Planarity) const {
  return IteratorP(new ScalarFieldIterator);
}

IndexP ScalarFieldBase::componenttype() const {
  return IndexP(new ScalarFieldIndex);
}

IndexP ScalarFieldBase::getIndex(const std::string &) const {
  return IndexP(new ScalarFieldIndex);
}

const std::string ScalarFieldBase::classname_("ScalarFieldBase");

//------------------------------

TwoVectorField::TwoVectorField(const std::string &nm)
  : Field(nm, 2),
    TwoVectorFieldBase(nm),
    CompoundField(nm, 2, new ThreeVectorField(nm+std::string("_z")))
{
}

const std::string TwoVectorField::classname_("TwoVectorField");

DegreeOfFreedom *TwoVectorFieldBase::operator()(const FuncNode *node, int comp)
  const
{
  assert(comp >= 0 && comp < 2);
  assert(node->hasField(*this));
  return node->doflist[node->fieldset.offset(this) + comp];
}

DegreeOfFreedom *
TwoVectorFieldBase::operator()(const ElementFuncNodeIterator &ei, int comp)
  const
{
  return operator()(ei.funcnode(), comp);
}

OutputValue TwoVectorFieldBase::newOutputValue() const {
  return OutputValue(new VectorOutputVal(2));
}

OutputValue TwoVectorFieldBase::output(const FuncNode &node) const {
  VectorOutputVal *vov = new VectorOutputVal(2);
  OutputValue ov(vov);
  if(node.hasField(*this)) {
    (*vov)[0] = operator()(node, 0)->value();
    (*vov)[1] = operator()(node, 1)->value();
  }
  return ov;
}

OutputValue TwoVectorFieldBase::output(const ElementFuncNodeIterator &node)
  const 
{
  VectorOutputVal *vov = new VectorOutputVal(2);
  OutputValue ov(vov);
  if(node.hasField(*this)) {
    (*vov)[0] = operator()(node, 0)->value();
    (*vov)[1] = operator()(node, 1)->value();
  }
  return ov;
}

void TwoVectorFieldBase::setValueFromOutputValue(const FuncNode &node,
						 const OutputValue *ov)
{
  assert(node.hasField(*this));
  const VectorOutputVal &vov = 
    dynamic_cast<const VectorOutputVal&>(ov->valueRef());
  (*this)(node, 0)->value() = vov[0];
  (*this)(node, 1)->value() = vov[1];
}

// The TwoVectorHelper class allows TwoVectorField(FuncNode) to act like a
// Coord.
TwoVectorHelper TwoVectorFieldBase::operator()(const FuncNode *node) const {
  assert(node->hasField(*this));
  int idx = node->fieldset.offset(this);
  return TwoVectorHelper(node->doflist[idx], node->doflist[idx+1]);
}

TwoVectorHelper
TwoVectorFieldBase::operator()(const ElementFuncNodeIterator &ei) const {
  return operator()(ei.funcnode());
}

// field(node) = coord
TwoVectorHelper &TwoVectorHelper::operator=(const Coord &v) {
  x->value() = v(0);
  y->value() = v(1);
  return *this;
}

// coord = field(node)
TwoVectorHelper::operator const Coord() const {
  return Coord(x->value(), y->value());
}

IteratorP TwoVectorFieldBase::iterator(Planarity) const {
  return IteratorP(new VectorFieldIterator(0, 2));
}

IndexP TwoVectorFieldBase::componenttype() const {
  return IndexP(new VectorFieldIndex);
}

IndexP TwoVectorFieldBase::getIndex(const std::string &str) const {
  return IndexP(new VectorFieldIndex(str[0] - 'x'));
}

const std::string TwoVectorFieldBase::classname_("TwoVectorFieldBase");

//-------------------------------


const std::string VectorFieldBase::classname_("VectorFieldBase");

DegreeOfFreedom *VectorFieldBase::operator()(const FuncNode *node, int comp)
  const
{
  assert(comp >= 0 && comp < dim);
  assert(node->hasField(*this));
  return node->doflist[node->fieldset.offset(this) + comp];
}

DegreeOfFreedom *VectorFieldBase::operator()(const ElementFuncNodeIterator &ei,
					     int comp)
  const
{
  return operator()(ei.funcnode(), comp);
}

OutputValue VectorFieldBase::newOutputValue() const {
  return OutputValue(new VectorOutputVal(dim));
}

OutputValue VectorFieldBase::output(const FuncNode &node) const {
  VectorOutputVal *vov = new VectorOutputVal(dim);
  OutputValue ov(vov);
  if(node.hasField(*this))
    for(int i=0; i<dim; i++)
      (*vov)[i] = operator()(node, i)->value();
  return ov;
}

OutputValue VectorFieldBase::output(const ElementFuncNodeIterator &node) const {
  VectorOutputVal *vov = new VectorOutputVal(dim);
  OutputValue ov(vov);
  if(node.hasField(*this))
    for(int i=0; i<dim; i++)
      (*vov)[i] = operator()(node, i)->value();
  return ov;
}

void VectorFieldBase::setValueFromOutputValue(const FuncNode &node,
					      const OutputValue *ov)
{
  assert(node.hasField(*this));
  const VectorOutputVal &vov = 
    dynamic_cast<const VectorOutputVal&>(ov->valueRef());
  for(int i=0; i<dim; i++)
    (*this)(node, i)->value() = vov[i];
}

// The VectorHelper class allows VectorField(FuncNode) to act like a Vector.
VectorHelper VectorFieldBase::operator()(const FuncNode *node) const {
  assert(node->hasField(*this));
  return VectorHelper(&node->doflist[node->fieldset.offset(this)], dim);
}

VectorHelper VectorFieldBase::operator()(const ElementFuncNodeIterator &ei)
  const
{
  return operator()(ei.funcnode());
}

// field(node) = vector
VectorHelper &VectorHelper::operator=(const VECTOR_D &v) {
#ifdef DEBUG
  assert(v.dim() == ndof);
#endif
  for(unsigned int i=0; i<ndof; i++)
    x[i]->value() = v[i];
  return *this;
}

// vector = field(node)
VectorHelper::operator const VECTOR_D() const {
  VECTOR_D v(ndof);
  for(unsigned int i=0; i<ndof; i++)
    v[i] = x[i]->value();
  return v;
}

IteratorP VectorFieldBase::iterator(Planarity planarity) const {
  int mindim = (planarity == OUT_OF_PLANE ? 2 : 0);
  int maxdim = (planarity == IN_PLANE ? 2 : ndof());
  return IteratorP(new VectorFieldIterator(mindim, maxdim));
}

IndexP VectorFieldBase::componenttype() const {
  return IndexP(new VectorFieldIndex);
}

IndexP VectorFieldBase::getIndex(const std::string &str) const {
  return IndexP(new VectorFieldIndex(str[0] - 'x'));
}


const std::string ThreeVectorField::classname_("ThreeVectorField");
