// -*- C++ -*-
// $RCSfile: linearizedsystem.C,v $
// $Revision: 1.5 $
// $Author: rlua $
// $Date: 2007-04-29 22:49:55 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

// The StiffnessMatrix class stores the components of the global
// stiffness matrix.  The components are accessed by equation number
// (as returned by Node::eqno() and by degree of freedom index (as
// returned by DegreeOfFreedom::dofindex()).  The StiffnessMatrix can
// compute submatrices corresponding to the fixed and free equations
// and degrees of freedom.  It returns these as compressed sparse row
// (CSRmatrix) matrices.

#include <oofconfig.h>
#include <iostream>
#include <vector>
#include "common/ooferror.h"
#include "common/tostring.h"
#include "common/trace.h"
#include "engine/csrmatrix.h"
#include "engine/linearizedsystem.h"
#include "engine/sparselink.h"
#include "engine/csubproblem.h"

LinearizedSystem::LinearizedSystem(CSubProblem *subp)
  : subproblem(subp), stiffness(0), kinetic(0), dynamic(0), 
    body_rhs(0), boundary_rhs(0),
    rows(0), cols(0), initialized(false), 
    has_kinetics(false), has_dynamics(false),
    rowgroup(0), colgroup(0)
{
}

void LinearizedSystem::initialize(int nr, int nc) {
  if (initialized) { 
    throw ErrSetupError("Attempt to reinitialize linear system.");
  }
  rows = nr;
  cols = nc;
  stiffness = new SparseLinkMatrix<double>(nr,nc);
  body_rhs = new VECTOR_D(nr, 0.0);
  boundary_rhs = new VECTOR_D(nr, 0.0);
  rowgroup.resize(nr, 0);
  colgroup.resize(nc, 0);
  initialized = true;
}

void LinearizedSystem::set_kinetic() { 
  if (!initialized) {
    throw ErrSetupError("Initialize LinearizedSystem before setting kinetics.");
  }
  has_kinetics = true;
  kinetic = new SparseLinkMatrix<double>(rows,cols);
}

void LinearizedSystem::set_dynamic() {
  if (!initialized) {
    throw ErrSetupError("Initialize LinearizedSystem before setting kinetics.");
  }
  has_dynamics = true;
  dynamic = new SparseLinkMatrix<double>(rows,cols);
}

LinearizedSystem::~LinearizedSystem() {
  if (initialized) {
    delete stiffness;
    delete body_rhs;
    delete boundary_rhs;
  }
  if (has_kinetics)
    delete kinetic;
  if (has_dynamics) 
    delete dynamic;
}

void LinearizedSystem::clear_brhs() {
    boundary_rhs->clear();
}


void LinearizedSystem::clear() {
  if (initialized) {
    delete stiffness;
    delete body_rhs;
    delete boundary_rhs;
    rows = 0;
    cols = 0;
    rowgroup.clear();
    colgroup.clear();
    initialized = false;
  }
  if (has_kinetics) {
    delete kinetic;
    has_kinetics = false;
  }
  if (has_dynamics) {
    delete dynamic;
    has_dynamics = false;
  }
}

// A row or column is nonempty if it's nonempty in any of the
// matrices.
bool LinearizedSystem::is_nonempty_row(int r) const {
  return stiffness->is_nonempty_row(r) ||
    (has_kinetics && kinetic->is_nonempty_row(r)) ||
    (has_dynamics && dynamic->is_nonempty_row(r));
//   bool res = stiffness->is_nonempty_row(r);
//   if (has_kinetics)
//     res = res || kinetic->is_nonempty_row(r);
//   if (has_dynamics)
//     res = res || dynamic->is_nonempty_row(r);
//   return res;
}

bool LinearizedSystem::is_nonempty_col(int c) const {
  return stiffness->is_nonempty_col(c) ||
    (has_kinetics && kinetic->is_nonempty_col(c)) ||
    (has_dynamics && dynamic->is_nonempty_col(c));
//   bool res = stiffness->is_nonempty_col(c);
//   if (has_kinetics)
//     res = res || kinetic->is_nonempty_col(c);
//   if (has_dynamics)
//     res = res || dynamic->is_nonempty_col(c);
//   return res;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

CSRmatrix 
LinearizedSystem::stiffness_submatrix(const std::vector<int> &rowmap,
				      const std::vector<int> &colmap) const
{
//   int rmax = 0, cmax=0;
//   for(int i=0; i<rowmap.size(); i++)
//     if(rowmap[i] > rmax) rmax = rowmap[i];
//   for(int i=0; i<colmap.size(); i++)
//     if(colmap[i] > cmax) cmax = colmap[i];
//   std::cerr << "StiffnessMatrix::submatrix: " << this << " " << nrows() << "x" << ncols()
// 	    << " rmax=" << rmax << " cmax=" << cmax << std::endl;
  return CSRmatrix(*stiffness, rowmap, colmap);
}

CSRmatrix 
LinearizedSystem::kinetic_submatrix(const std::vector<int> &rowmap,
				      const std::vector<int> &colmap) const
{
  return CSRmatrix(*kinetic, rowmap, colmap);
}


CSRmatrix 
LinearizedSystem::dynamic_submatrix(const std::vector<int> &rowmap,
				      const std::vector<int> &colmap) const
{
  return CSRmatrix(*dynamic, rowmap, colmap);
}



std::vector<int> LinearizedSystem::subgroup(const std::vector<int> groupset,
					    const std::vector<int> &map) const
{  
  // The "groupset" is a list, indexed by DOF, which is -1 if the DOF
  // is in-plane, and is equal to the node number (>=0) if the DOF is
  // out of plane.  The "map" is a list, also indexed by DOF, which is
  // -1 if the DOF will not be included in the submatrix, and
  // otherwise is equal to the column number of the submatrix
  // corresponding to this DOF.  Because of floating BCs, there can be
  // duplicates in the map, but these duplicates will always be
  // in-plane.  Likewise for Equations/rows.
  //   The output of this routine should be a third list, indexed by
  // column of the submatrix, with values of -1 for in-plane DOFs,
  // and the node number for out-of-plane DOFs,
  int subgroup_list_size = 0;
  for(std::vector<int>::size_type i=0; i<map.size(); i++) {
    if (map[i] >= subgroup_list_size) subgroup_list_size = map[i];
  }
  subgroup_list_size++;

  // Invert the map -- in case of duplicates, retain only the DOF/eqn
  // number of the first occurrence.  This is what the submatrix
  // extraction code does.
  std::vector<int> map_inverse(subgroup_list_size,-1);
  for(std::vector<int>::size_type i=0; i<map.size(); i++) {
    if (map[i]>=0) {
      if (map_inverse[map[i]] == -1) { // First occurrence only happens here.
	map_inverse[map[i]]=i;
      }
    }
  }

  std::vector<int> res(subgroup_list_size);

  for(std::vector<int>::size_type i=0; i<res.size(); i++) {
    res[i] = groupset[map_inverse[i]];
  }

  return res;
}

// Extract the subset of the rowgroups corresponding to the given rowmap.
std::vector<int> LinearizedSystem::rowgroups(const std::vector<int> &rowmap)
  const 
{
  return subgroup(rowgroup, rowmap);
}


// Extract the subset of colgroups corresponding to the given map.
std::vector<int> LinearizedSystem::colgroups(const std::vector<int> &colmap)
  const 
{
  return subgroup(colgroup, colmap);
}
