# -*- python -*-
# $RCSfile: rationalsharp.py,v $
# $Revision: 1.24 $
# $Author: langer $
# $Date: 2005-05-13 18:44:06 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from oof2.common import debug
from oof2.common import registeredclass
from oof2.common import utils
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump
from oof2.engine import rationalize
from oof2.engine import rationalizefunctions
import math

Rationalizer = rationalize.Rationalizer
removeBadTriangle = rationalizefunctions.removeBadTriangle

class RemoveBadTriangle(Rationalizer):
    #
    #             C
    #            /|
    #           / |  If an interior angle A-C-B is too small,
    #          /  |  it's better to remove this triangle.
    #         /   |  
    #        /____|  This case involves a "merging" process.
    #        A    B  (merging of two nodes, A and B)
    #
    def __init__(self, acute_angle, obtuse_angle):
        self.acute_angle = acute_angle
        self.obtuse_angle = obtuse_angle
        self.acute = math.cos(acute_angle*math.pi/180.0)
        self.obtuse = math.cos(obtuse_angle*math.pi/180.0)

    def findAndFix(self, skel, element):
        if element.nnodes() == 4:
            return []
        
        obtuse_angle = None
        acuteCheck = None
        for i in range(3):
            if element.cosCornerAngle(i) <= self.obtuse:
                obtuse_angle = i
        for i in range(3):
            if element.cosCornerAngle(i) >= self.acute:
                acuteCheck = i
                break

        # See if there is any violation ....
        if obtuse_angle is None and acuteCheck is None:
            return []

        # Let's fix them.
        if obtuse_angle is None:
            obtuse_angle = element.getBiggestAngle()
        acute_angles = [(obtuse_angle+1)%3, (obtuse_angle+2)%3]
        return self.fix(skel, element, obtuse_angle, acute_angles)

    def fixAll(self, skel, element):
        if element.nnodes() == 4:
            return []
        obtuse_angle = element.getBiggestAngle()
        acute_angles = [(obtuse_angle+1)%3, (obtuse_angle+2)%3]
        return self.fix(skel, element, obtuse_angle, acute_angles)

    def fix(self, skel, element, obtuse_angle, acute_angles):
        return removeBadTriangle(skel, element, obtuse_angle, acute_angles)

registeredclass.Registration(
    'Remove Bad Triangles',
    Rationalizer,
    RemoveBadTriangle,
    gerund = 'removing bad triangles',
    ordering=20000,                     # do this last!
    params=[
    parameter.FloatRangeParameter('acute_angle', (0.0, 45.0, 0.5),
                                  value = 15.0,
                                  tip = 'Minimum acceptable acute interior angle, in degrees'),
    parameter.FloatRangeParameter('obtuse_angle', (90.0, 180.0, 1.0),
                                  value = 150.0,
                                  tip = 'Maximum acceptable obtuse interior angle, in degrees')],
    tip = 'Remove triangles with extreme interior angles.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/ration_sharp.xml'))
